@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 2005 Rob Braun
+ * Copyright (c) 2005-2008 Rob Braun
+ * Portions Copyright (c) 2012 Kyle J. McKay.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -28,7 +29,7 @@
  */
 /*
  * 03-Apr-2005
- * DRI: Rob Braun <bbraun@opendarwin.org>
+ * DRI: Rob Braun <bbraun@synack.net>
  */
 /*
  * Portions Copyright 2006, Apple Computer, Inc.
@@ -37,6 +38,7 @@
 
 #define _FILE_OFFSET_BITS 64
 
+#include "config.h"
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -51,13 +53,13 @@
 #include <libxml/xmlwriter.h>
 #include <libxml/xmlreader.h>
 #include <libxml/xmlstring.h>
-#include "config.h"
 #ifndef HAVE_ASPRINTF
 #include "asprintf.h"
 #endif
 #include "xar.h"
 #include "filetree.h"
 #include "archive.h"
+#include "hash.h"
 #include "signature.h"
 #include "arcmod.h"
 #include "io.h"
@@ -65,6 +67,10 @@
 #include "subdoc.h"
 #include "darwinattr.h"
 
+#define _XAR_LIB_VERSION2(x) #x
+#define _XAR_LIB_VERSION1(x) _XAR_LIB_VERSION2(x)
+#define XAR_LIB_VERSION() _XAR_LIB_VERSION1(XAR_VERSION_NUM)
+
 #ifndef O_EXLOCK
 #define O_EXLOCK 0
 #endif
@@ -79,6 +85,10 @@
 #define LONG_MIN INT32_MIN
 #endif
 
+#if LIBXML_VERSION < 20618
+#define xmlDictCleanup()	/* function doesn't exist in older API */
+#endif
+
 static int32_t xar_unserialize(xar_t x);
 void xar_serialize(xar_t x, const char *file);
 
@@ -92,7 +102,7 @@ static xar_t xar_new() {
 	ret = malloc(sizeof(struct __xar_t));
 	if(!ret) return NULL;
 	memset(XAR(ret), 0, sizeof(struct __xar_t));
-	XAR(ret)->readbuf_len = 4096;
+	XAR(ret)->readbuf_len = XAR_DEFAULT_BUFFER_SIZE;
 	XAR(ret)->readbuf = malloc(XAR(ret)->readbuf_len);
 	if(!XAR(ret)->readbuf) {
 		free((void *)ret);
@@ -107,7 +117,13 @@ static xar_t xar_new() {
 	XAR(ret)->link_hash = xmlHashCreate(0);
 	XAR(ret)->csum_hash = xmlHashCreate(0);
 	XAR(ret)->subdocs = NULL;
-	
+	XAR(ret)->toc_ctx = EVP_MD_CTX_create();
+	if(!XAR(ret)->toc_ctx) {
+		free(XAR(ret)->readbuf);
+		free((void *)ret);
+		return NULL;
+	}
+
 	return ret;
 }
 
@@ -118,7 +134,7 @@ static xar_t xar_new() {
  */
 static int32_t xar_parse_header(xar_t x) {
 	ssize_t r;
-	int off = 0;
+	int off;
 	int sz2read = 0;
 
 	/* read just the magic, verify it, read the header length,
@@ -128,9 +144,9 @@ static int32_t xar_parse_header(xar_t x) {
 	 * if the recorded header length is greater than the 
 	 * expected header length.
 	 */
-	r = xar_read_fd(XAR(x)->fd, (char *)&XAR(x)->header.magic+off, sizeof(XAR(x)->header.magic)-off);
+	r = xar_read_fd(XAR(x)->fd, (char *)&XAR(x)->header.magic, sizeof(XAR(x)->header.magic));
 	if ( r == -1 )
-		return r;
+		return -1;
 
 	/* Verify the header.  If the header doesn't match, exit without
 	 * attempting to read any more.
@@ -141,33 +157,48 @@ static int32_t xar_parse_header(xar_t x) {
 		return -1;
 	}
 
-	r = xar_read_fd(XAR(x)->fd, (char *)&XAR(x)->header.size+off, sizeof(XAR(x)->header.size)-off);
+	r = xar_read_fd(XAR(x)->fd, (char *)&XAR(x)->header.size, sizeof(XAR(x)->header.size));
 	if ( r == -1 )
-		return r;
+		return -1;
 
 	XAR(x)->header.size = ntohs(XAR(x)->header.size);
 
-	if( XAR(x)->header.size > sizeof(xar_header_t) )
-		sz2read = sizeof(xar_header_t);
+	if( XAR(x)->header.size > sizeof(xar_header_ex_t) )
+		sz2read = (int)sizeof(xar_header_ex_t);
 	else
 		sz2read = XAR(x)->header.size;
+	if ( XAR(x)->header.size < sizeof(xar_header_t) )
+		return -1;
 
-	off = sizeof(XAR(x)->header.magic) + sizeof(XAR(x)->header.size);
-	r = xar_read_fd(XAR(x)->fd, ((char *)&XAR(x)->header)+off, sizeof(xar_header_t)-off);
+	off = (int)(sizeof(XAR(x)->header.magic) + sizeof(XAR(x)->header.size));
+	r = xar_read_fd(XAR(x)->fd, ((char *)&XAR(x)->header)+off, sz2read-off);
 	if ( r == -1 )
-		return r;
+		return -1;
 
 	XAR(x)->header.version = ntohs(XAR(x)->header.version);
 	XAR(x)->header.toc_length_compressed = xar_ntoh64(XAR(x)->header.toc_length_compressed);
 	XAR(x)->header.toc_length_uncompressed = xar_ntoh64(XAR(x)->header.toc_length_uncompressed);
 	XAR(x)->header.cksum_alg = ntohl(XAR(x)->header.cksum_alg);
 
+	if (XAR(x)->header.cksum_alg == XAR_CKSUM_OTHER) {
+		/* Make sure toc_cksum_name is Nul terminated and included in header length */
+		size_t max_search = XAR(x)->header.size;
+		if (max_search > sizeof(xar_header_ex_t))
+			max_search = sizeof(xar_header_ex_t);
+		max_search -= sizeof(xar_header_t);
+		if (XAR(x)->header.size < sizeof(xar_header_t) + 4 || (XAR(x)->header.size & 0x3) != 0)
+			return -1; /* Does not include a name or not multiple of 4 */
+		if (!memchr(XAR(x)->header.toc_cksum_name, 0, max_search))
+			return -1; /* No Nul terminated string present for toc_cksum_name within header.size bytes */
+	}
+
 	off = XAR(x)->header.size - sz2read;
 	if( off > 0 )
-		r = lseek(XAR(x)->fd, (off_t)off, SEEK_CUR);
+		r = (int)lseek(XAR(x)->fd, (off_t)off, SEEK_CUR);
 
-	if ( (r == -1) && (errno != ESPIPE) )
-		/* Some fatal error here perhaps? */ ;
+	if ( (r == -1) && (errno != ESPIPE) ) {
+		/* Some fatal error here perhaps? */
+	}
 
 	return 0;
 }
@@ -187,6 +218,10 @@ xar_t xar_open(const char *file, int32_t flags) {
 	if( !file )
 		file = "-";
 	XAR(ret)->filename = strdup(file);
+	if (!*file && flags) {
+		XAR(ret)->fd = XAR(ret)->heap_fd = -2;
+		return ret;
+	}
 	OpenSSL_add_all_digests();
 	if( flags ) {
 		char *tmp1, *tmp2, *tmp3, *tmp4;
@@ -196,7 +231,12 @@ xar_t xar_open(const char *file, int32_t flags) {
 		/* Create the heap file in the directory which will contain
 		 * the target archive.  /tmp or elsewhere may fill up.
 		 */
-		asprintf(&tmp4, "%s/xar.heap.XXXXXX", tmp3);
+		if (asprintf(&tmp4, "%s/xar.heap.XXXXXX", tmp3) == -1) {
+			free(XAR(ret)->dirname);
+			free((void *)XAR(ret)->filename);
+			free(XAR(ret));
+			return NULL;
+		}
 		free(tmp1);
 		if( strcmp(file, "-") == 0 )
 			XAR(ret)->fd = 1;
@@ -209,6 +249,8 @@ xar_t xar_open(const char *file, int32_t flags) {
 		XAR(ret)->heap_fd = mkstemp(tmp4);
 		if( XAR(ret)->heap_fd < 0 ) {
 			close(XAR(ret)->fd);
+			free(XAR(ret)->dirname);
+			free((void *)XAR(ret)->filename);
 			free(XAR(ret));
 			return NULL;
 		}
@@ -219,6 +261,9 @@ xar_t xar_open(const char *file, int32_t flags) {
 
 		if( XAR(ret)->fd < 0 ) {
 			xar_close(ret);
+			free(XAR(ret)->dirname);
+			free((void *)XAR(ret)->filename);
+			free(XAR(ret));
 			return NULL;
 		}
 
@@ -231,10 +276,15 @@ xar_t xar_open(const char *file, int32_t flags) {
 		xar_opt_set(ret, XAR_OPT_COMPRESSION, XAR_OPT_VAL_GZIP);
 		xar_opt_set(ret, XAR_OPT_FILECKSUM, XAR_OPT_VAL_SHA1);
 	} else {
-		unsigned char toccksum[EVP_MAX_MD_SIZE];
-		unsigned char cval[EVP_MAX_MD_SIZE];
+		unsigned char toccksum[HASH_MAX_MD_SIZE];
+		unsigned char cval[HASH_MAX_MD_SIZE];
 		unsigned int tlen;
 		const EVP_MD *md;
+		int cksum_match;
+		const char *cksum_style;
+		const char *value;
+		uint64_t offset;
+		uint64_t length;
 
 		if( strcmp(file, "-") == 0 )
 			XAR(ret)->fd = 0;
@@ -259,18 +309,29 @@ xar_t xar_open(const char *file, int32_t flags) {
 		}
 
 		switch(XAR(ret)->header.cksum_alg) {
-		case XAR_CKSUM_NONE:
+		case XAR_CKSUM_MD5:
+			strncpy(XAR(ret)->header.toc_cksum_name, XAR_OPT_VAL_MD5, sizeof(XAR(ret)->header.toc_cksum_name));
 			break;
 		case XAR_CKSUM_SHA1:
-			XAR(ret)->docksum = 1;
-			md = EVP_get_digestbyname("sha1");
-			EVP_DigestInit(&XAR(ret)->toc_ctx, md);
+			strncpy(XAR(ret)->header.toc_cksum_name, XAR_OPT_VAL_SHA1, sizeof(XAR(ret)->header.toc_cksum_name));
+			break;
+		default:
+			/* nothing to do */;
+		};
+
+		switch(XAR(ret)->header.cksum_alg) {
+		case XAR_CKSUM_NONE:
 			break;
 		case XAR_CKSUM_MD5:
-			XAR(ret)->docksum = 1;
-			md = EVP_get_digestbyname("md5");
-			EVP_DigestInit(&XAR(ret)->toc_ctx, md);
-			break;
+		case XAR_CKSUM_SHA1:
+		case XAR_CKSUM_OTHER:
+			md = EVP_get_digestbyname(XAR(ret)->header.toc_cksum_name);
+			if (md) {
+				XAR(ret)->docksum = 1;
+				EVP_DigestInit_ex(XAR(ret)->toc_ctx, md, NULL);
+				break;
+			}
+			/* else fall through */
 		default:
 			fprintf(stderr, "Unknown hashing algorithm, skipping\n");
 			break;
@@ -281,10 +342,87 @@ xar_t xar_open(const char *file, int32_t flags) {
 			return NULL;
 		}
 
+		/* check for inconsistency between checksum style in header, 
+		 * and the one described in the TOC; otherwise, you can flip 
+		 * the header bit to XAR_CKSUM_NONE, and nothing will ever 
+		 * verify that the TOC matches the checksum stored in the 
+		 * heap, and the signature check will pass on a modified 
+		 * file! <rdar://problem/6134714>
+		 */
+		cksum_match = 0;
+		cksum_style = xar_attr_get(XAR_FILE(ret), "checksum", "style");
+		switch(XAR(ret)->header.cksum_alg) {
+			case XAR_CKSUM_NONE:
+				cksum_match = (cksum_style == NULL || strcmp(cksum_style, XAR_OPT_VAL_NONE) == 0);
+				break;
+			case XAR_CKSUM_MD5:
+			case XAR_CKSUM_SHA1:
+			case XAR_CKSUM_OTHER:
+				cksum_match = (cksum_style != NULL && strcmp(cksum_style, XAR(ret)->header.toc_cksum_name) == 0);
+				break;
+			default:
+				cksum_match = 0;
+				break;
+		}
+		if( !cksum_match ) {
+			fprintf(stderr, "Checksum style mismatch!\n");
+			xar_close(ret);
+			return NULL;
+		}
+		
+		/* also check for consistency between the checksum style and 
+		 * the existence (or not) of signatures: since the signature 
+		 * is signing the checksum, we must have a checksum to verify 
+		 * that the TOC has not been modified <rdar://problem/6134714>
+		 */
+		if( xar_signature_first(ret) != NULL && XAR(ret)->header.cksum_alg == XAR_CKSUM_NONE ) {
+			fprintf(stderr, "Checksum/signature mismatch!\n");
+			xar_close(ret);
+			return NULL;
+		}
+			
 		if( !XAR(ret)->docksum )
 			return ret;
 
-		EVP_DigestFinal(&XAR(ret)->toc_ctx, toccksum, &tlen);
+		EVP_DigestFinal_ex(XAR(ret)->toc_ctx, toccksum, &tlen);
+
+		offset = 0;
+		length = tlen;
+		if( xar_prop_get( XAR_FILE(ret) , "checksum/offset", &value) == 0 ) {
+			errno = 0;
+			offset = (uint64_t)strtoull(value, (char **)NULL, 10);
+			if( errno != 0 ) {
+				xar_close(ret);
+				return NULL;
+			}
+		} else if( xar_signature_first(ret) != NULL ) {
+			/* All archives that have a signature also specify the location
+			 * of the checksum.  If the location isn't specified, error out.
+			 */
+			xar_close(ret);
+			return NULL;
+		}
+
+		XAR(ret)->heap_offset = (off_t)(xar_get_heap_offset(ret) + offset);
+		if( lseek(XAR(ret)->fd, XAR(ret)->heap_offset, SEEK_SET) == -1 ) {
+			xar_close(ret);
+			return NULL;
+		}
+		if( xar_prop_get( XAR_FILE(ret) , "checksum/size", &value) == 0 ) {
+			errno = 0;
+			length = (uint64_t)strtoull(value, (char **)NULL, 10);
+			if( errno != 0 ) {
+				xar_close(ret);
+				return NULL;
+			}
+		} else if( xar_signature_first(ret) != NULL ) {
+			xar_close(ret);
+			return NULL;
+		}
+		if( length != tlen ) {
+			xar_close(ret);
+			return NULL;
+		}
 
 		xar_read_fd(XAR(ret)->fd, cval, tlen);
 		XAR(ret)->heap_offset += tlen;
@@ -298,6 +436,27 @@ xar_t xar_open(const char *file, int32_t flags) {
 	return ret;
 }
 
+/* xar_check_force_rfc6713
+ * x: the xar_t to check
+ * Summary: forces the rfcformat option on if the checksum type will be
+ *          XAR_CKSUM_OTHER.  Should be called just before first file is added.
+ */
+static void xar_check_force_rfc6713(xar_t x)
+{
+	char *tmpser;
+
+	if (XAR(x)->files != NULL)
+		return;
+	tmpser = (char *)xar_opt_get(x, XAR_OPT_TOCCKSUM);
+	if (!tmpser)
+		return;
+	if (strcmp(tmpser, XAR_OPT_VAL_NONE) == 0 ||
+		strcmp(tmpser, XAR_OPT_VAL_SHA1) == 0 ||
+		strcmp(tmpser, XAR_OPT_VAL_MD5) == 0)
+		return;
+	XAR(x)->rfcformat = 1;
+}
+
 /* xar_close
  * x: the xar_t to close
  * Summary: closes all open file descriptors, frees all
@@ -309,6 +468,9 @@ int xar_close(xar_t x) {
 	xar_file_t f;
 	int ret, retval = 0;
 
+	if (XAR(x)->heap_fd == -2)
+		goto CLOSE_BAIL;
+
 	/* If we're creating an archive */
 	if( XAR(x)->heap_fd != -1 ) {
 		char *tmpser;
@@ -317,11 +479,15 @@ int xar_close(xar_t x) {
 		long rsize, wsize;
 		z_stream zs;
 		uint64_t ungztoc, gztoc;
-		unsigned char chkstr[EVP_MAX_MD_SIZE];
+		unsigned char chkstr[HASH_MAX_MD_SIZE];
 		int tocfd;
 		char timestr[128];
 		struct tm tmptm;
 		time_t t;
+		uint32_t cksum_alg = XAR_CKSUM_NONE;
+		const char *opt;
+		size_t cnt;
+		ssize_t wcnt;
 
 		tmpser = (char *)xar_opt_get(x, XAR_OPT_TOCCKSUM);
 		/* If no checksum type is specified, default to sha1 */
@@ -329,42 +495,55 @@ int xar_close(xar_t x) {
 
 		if( (strcmp(tmpser, XAR_OPT_VAL_NONE) != 0) ) {
 			const EVP_MD *md;
-			xar_prop_set(XAR_FILE(x), "checksum", NULL);
-			if( strcmp(tmpser, XAR_OPT_VAL_SHA1) == 0 ) {
-				md = EVP_get_digestbyname("sha1");
-				EVP_DigestInit(&XAR(x)->toc_ctx, md);
-				XAR(x)->header.cksum_alg = htonl(XAR_CKSUM_SHA1);
-				xar_attr_set(XAR_FILE(x), "checksum", "style", XAR_OPT_VAL_SHA1);
-				xar_prop_set(XAR_FILE(x), "checksum/size", "20");
-			}
-			if( strcmp(tmpser, XAR_OPT_VAL_MD5) == 0 ) {
-				md = EVP_get_digestbyname("md5");
-				EVP_DigestInit(&XAR(x)->toc_ctx, md);
-				XAR(x)->header.cksum_alg = htonl(XAR_CKSUM_MD5);
-				xar_attr_set(XAR_FILE(x), "checksum", "style", XAR_OPT_VAL_MD5);
-				xar_prop_set(XAR_FILE(x), "checksum/size", "16");
-			}
+			const char *md_name;
+			char mdlenstr[32];
 
+			md = EVP_get_digestbyname(tmpser);
+			if( md == NULL ) return -1;
+			md_name = OBJ_nid2ln(EVP_MD_nid(md));
+			if( strcmp(md_name, XAR_OPT_VAL_SHA1) == 0 )
+				cksum_alg = XAR_CKSUM_SHA1;
+			else if( strcmp(md_name, XAR_OPT_VAL_MD5) == 0 )
+				cksum_alg = XAR_CKSUM_MD5;
+			else if( strlen(md_name) + 1 > sizeof(XAR(x)->header.toc_cksum_name) ) {
+				retval = -1;
+				goto CLOSE_BAIL;
+			}
+			else {
+				cksum_alg = XAR_CKSUM_OTHER;
+				strncpy(XAR(x)->header.toc_cksum_name, md_name, sizeof(XAR(x)->header.toc_cksum_name));
+			}
+			xar_prop_set(XAR_FILE(x), "checksum", NULL);
+			EVP_DigestInit_ex(XAR(x)->toc_ctx, md, NULL);
+			xar_attr_set(XAR_FILE(x), "checksum", "style", md_name);
+			sprintf(mdlenstr, "%d", EVP_MD_size(md));
+			xar_prop_set(XAR_FILE(x), "checksum/size", mdlenstr);
 			xar_prop_set(XAR_FILE(x), "checksum/offset", "0");
 			XAR(x)->docksum = 1;
 		} else {
 			XAR(x)->docksum = 0;
-			XAR(x)->header.cksum_alg = XAR_CKSUM_NONE;
 		}
+		XAR(x)->header.cksum_alg = htonl(cksum_alg);
 
 		t = time(NULL);
 		gmtime_r(&t, &tmptm);
 		memset(timestr, 0, sizeof(timestr));
-		strftime(timestr, sizeof(timestr), "%FT%T", &tmptm);
+		strftime(timestr, sizeof(timestr), "%Y-%m-%dT%H:%M:%SZ", &tmptm);
 		xar_prop_set(XAR_FILE(x), "creation-time", timestr);
 
 		/* serialize the toc to a tmp file */
-		asprintf(&tmpser, "%s/xar.toc.XXXXXX", XAR(x)->dirname);
+		if (asprintf(&tmpser, "%s/xar.toc.XXXXXX", XAR(x)->dirname) == -1) {
+			retval = -1;
+			goto CLOSE_BAIL;
+		}
 		fd = mkstemp(tmpser);
 		xar_serialize(x, tmpser);
 		unlink(tmpser);
 		free(tmpser);
-		asprintf(&tmpser, "%s/xar.toc.XXXXXX", XAR(x)->dirname);
+		if (asprintf(&tmpser, "%s/xar.toc.XXXXXX", XAR(x)->dirname) == -1) {
+			retval = -1;
+			goto CLOSE_BAIL;
+		}
 		tocfd = mkstemp(tmpser);
 		unlink(tmpser);
 		free(tmpser);
@@ -373,15 +552,17 @@ int xar_close(xar_t x) {
 		/* read the toc from the tmp file, compress it, and write it
 	 	* out to the archive.
 	 	*/
-		rsize = wsize = 4096;
-		const char * opt = xar_opt_get(x, XAR_OPT_RSIZE);
+		rsize = wsize = XAR_DEFAULT_BUFFER_SIZE;
+		opt = xar_opt_get(x, XAR_OPT_RSIZE);
 		if ( opt ) {
 		  rsize = strtol(opt, NULL, 0);
 		  if ( ((rsize == LONG_MAX) || (rsize == LONG_MIN)) && (errno == ERANGE) ) {
 		    rsize = wsize;
 		  }
+		  if (rsize < XAR_MINIMUM_BUFFER_SIZE)
+		    rsize = XAR_MINIMUM_BUFFER_SIZE;
 		}
-		
+
 		rbuf = malloc(rsize);
 		if( !rbuf ) {
 			retval = -1;
@@ -395,7 +576,7 @@ int xar_close(xar_t x) {
 		ungztoc = gztoc = 0;
 	
 		while(1) {
-			r = read(fd, rbuf, rsize);
+			r = (int)read(fd, rbuf, rsize);
 			if( (r < 0) && (errno == EINTR) )
 				continue;
 			if( r == 0 )
@@ -415,17 +596,17 @@ int xar_close(xar_t x) {
 				wsize *= 2;
 				wbuf = realloc(wbuf, wsize);
 
-				zs.next_out = wbuf + off;
-				zs.avail_out = wsize - off;
+				zs.next_out = ((unsigned char *)wbuf) + off;
+				zs.avail_out = (unsigned)wsize - off;
 
 				ret = deflate(&zs, Z_SYNC_FLUSH);
-				off = wsize - zs.avail_out;
+				off = (int)(wsize - zs.avail_out);
 			}
 	
 			wbytes = off;
 			off = 0;
 			do {
-				r = write(tocfd, wbuf+off, wbytes-off);
+				r = (int)write(tocfd, ((char *)wbuf)+off, wbytes-off);
 				if( (r < 0) && (errno == EINTR) )
 					continue;
 				if( r < 0 ) {
@@ -435,7 +616,7 @@ int xar_close(xar_t x) {
 					goto CLOSEEND;
 				}
 				if( XAR(x)->docksum )
-					EVP_DigestUpdate(&XAR(x)->toc_ctx, wbuf+off, r);
+					EVP_DigestUpdate(XAR(x)->toc_ctx, ((char*)wbuf)+off, r);
 				off += r;
 				gztoc += r;
 			} while( off < wbytes );
@@ -445,29 +626,39 @@ int xar_close(xar_t x) {
 		zs.next_in = NULL;
 		zs.avail_in = 0;
 		zs.next_out = wbuf;
-		zs.avail_out = wsize;
+		zs.avail_out = (unsigned)wsize;
 
 		deflate(&zs, Z_FINISH);
-		r = write(tocfd, wbuf, wsize - zs.avail_out);
+		r = (int)write(tocfd, wbuf, wsize - zs.avail_out);
 		gztoc += r;
 		if( XAR(x)->docksum )
-			EVP_DigestUpdate(&XAR(x)->toc_ctx, wbuf, r);
+			EVP_DigestUpdate(XAR(x)->toc_ctx, wbuf, r);
 		
 		deflateEnd(&zs);
 
 		/* populate the header and write it out */
 		XAR(x)->header.magic = htonl(XAR_HEADER_MAGIC);
-		XAR(x)->header.size = ntohs(sizeof(xar_header_t));
+		if (cksum_alg == XAR_CKSUM_OTHER)
+			XAR(x)->header.size = ntohs(sizeof(xar_header_ex_t));
+		else
+			XAR(x)->header.size = ntohs(sizeof(xar_header_t));
 		XAR(x)->header.version = ntohs(1);
 		XAR(x)->header.toc_length_uncompressed = xar_ntoh64(ungztoc);
 		XAR(x)->header.toc_length_compressed = xar_ntoh64(gztoc);
 
-		write(XAR(x)->fd, &XAR(x)->header, sizeof(xar_header_t));
+		cnt = cksum_alg == XAR_CKSUM_OTHER ? sizeof(xar_header_ex_t) : sizeof(xar_header_t);
+		wcnt = write(XAR(x)->fd, &XAR(x)->header, cnt);
+		if (wcnt < 0 || wcnt != (ssize_t)cnt) {
+			xar_err_new(x);
+			xar_err_set_string(x, "Error writing xar archive header");
+			retval = -1;
+			goto CLOSEEND;
+		}
 
 		/* Copy the temp compressed toc file into the file */
 		lseek(tocfd, (off_t)0, SEEK_SET);
 		while(1) {
-			r = read(tocfd, rbuf, rsize);
+			r = (int)read(tocfd, rbuf, rsize);
 			if( (r < 0) && (errno == EINTR) )
 				continue;
 			if( r == 0 )
@@ -476,7 +667,7 @@ int xar_close(xar_t x) {
 			wbytes = r;
 			off = 0;
 			do {
-				r = write(XAR(x)->fd, rbuf+off, wbytes-off);
+				r = (int)write(XAR(x)->fd, ((char *)rbuf)+off, wbytes-off);
 				if( (r < 0) && (errno == EINTR) )
 					continue;
 				if( r < 0 ) {
@@ -490,13 +681,23 @@ int xar_close(xar_t x) {
 			} while( off < wbytes );
 		}
 
+		/* start counting any digest/signature bytes written */
+		rbytes = 0;
+
 		if( XAR(x)->docksum ) {
 			unsigned int l = r;
 			
 			memset(chkstr, 0, sizeof(chkstr));
-			EVP_DigestFinal(&XAR(x)->toc_ctx, chkstr, &l);
+			EVP_DigestFinal_ex(XAR(x)->toc_ctx, chkstr, &l);
 			r = l;
-			write(XAR(x)->fd, chkstr, r);
+			wcnt = write(XAR(x)->fd, chkstr, r);
+			if (wcnt < 0 || wcnt != (ssize_t)r) {
+				xar_err_new(x);
+				xar_err_set_string(x, "Error writing xar archive checksum");
+				retval = -1;
+				goto CLOSEEND;
+			}
+			rbytes += r;
 		}
 
 		/* If there are any signatures, get the signed data a sign it */
@@ -517,15 +718,22 @@ int xar_close(xar_t x) {
 					goto CLOSE_BAIL;					
 				}
 				
-				if( signed_len != XAR_SIGNATURE(sig)->len ){
+				if( signed_len != (uint32_t)XAR_SIGNATURE(sig)->len ){
 					fprintf(stderr, "Signed data not the proper length.  %i should be %i.\n",signed_len,XAR_SIGNATURE(sig)->len);
 					retval = -1;
 					goto CLOSE_BAIL;										
 				}
 				
 				/* Write the signed data to the heap */
-				write(XAR(x)->fd, signed_data,XAR_SIGNATURE(sig)->len);
-				
+				wcnt = write(XAR(x)->fd, signed_data, (size_t)XAR_SIGNATURE(sig)->len);
+				if (wcnt < 0 || wcnt != (ssize_t)XAR_SIGNATURE(sig)->len) {
+					xar_err_new(x);
+					xar_err_set_string(x, "Error writing xar archive signature data");
+					retval = -1;
+					goto CLOSEEND;
+				}
+				rbytes += XAR_SIGNATURE(sig)->len;
+
 				free(signed_data);
 			}
 			
@@ -538,12 +746,13 @@ int xar_close(xar_t x) {
 			fprintf(stderr, "Error lseeking to offset 0: %s\n", strerror(errno));
 			exit(1);
 		}
-		rbytes = 0;
+		/* XAR(x)->heap_len includes any digest/signatures but the heap file does not and at this
+                 * point rbytes reflects the total byte count of any digest/signatures that are present */
 		while(1) {
 			if( (XAR(x)->heap_len - rbytes) < rsize )
-				rsize = XAR(x)->heap_len - rbytes;
+				rsize = (long)(XAR(x)->heap_len - rbytes);
 
-			r = read(XAR(x)->heap_fd, rbuf, rsize);
+			r = (int)read(XAR(x)->heap_fd, rbuf, rsize);
 			if( (r < 0 ) && (errno == EINTR) )
 				continue;
 			if( r == 0 )
@@ -553,7 +762,7 @@ int xar_close(xar_t x) {
 			wbytes = r;
 			off = 0;
 			do {
-				r = write(XAR(x)->fd, rbuf+off, wbytes);
+				r = (int)write(XAR(x)->fd, ((char *)rbuf)+off, wbytes);
 				if( (r < 0 ) && (errno == EINTR) )
 					continue;
 				if( r < 0 ) {
@@ -602,12 +811,14 @@ CLOSE_BAIL:
 	xmlHashFree(XAR(x)->ino_hash, NULL);
 	xmlHashFree(XAR(x)->link_hash, NULL);
 	xmlHashFree(XAR(x)->csum_hash, NULL);
-	close(XAR(x)->fd);
+	if (XAR(x)->fd >= 0)
+		close(XAR(x)->fd);
 	if( XAR(x)->heap_fd >= 0 )
 		close(XAR(x)->heap_fd);
 	free((char *)XAR(x)->filename);
 	free((char *)XAR(x)->dirname);
 	free(XAR(x)->readbuf);
+	EVP_MD_CTX_destroy(XAR(x)->toc_ctx);
 	free((void *)x);
 
 	return retval;
@@ -617,9 +828,15 @@ CLOSE_BAIL:
  * x: archive to get the option from
  * option: name of the option
  * Returns: a pointer to the value of the option
+ * In the case of more than one option with the same name, this will
+ * return the first match.
  */
 const char *xar_opt_get(xar_t x, const char *option) {
 	xar_attr_t i;
+	if (!x || !option)
+		return NULL;
+	if (strcmp(option, XAR_OPT_XARLIBVERSION) == 0)
+		return XAR_LIB_VERSION();
 	for(i = XAR(x)->attrs; i && XAR_ATTR(i)->next; i = XAR_ATTR(i)->next) {
 		if(strcmp(XAR_ATTR(i)->key, option)==0)
 			return XAR_ATTR(i)->value;
@@ -636,26 +853,65 @@ const char *xar_opt_get(xar_t x, const char *option) {
  * Returns: 0 for sucess, -1 for failure
  */
 int32_t xar_opt_set(xar_t x, const char *option, const char *value) {
-	xar_attr_t i, a;
+	xar_attr_t a;
 
+	if (!x || !option)
+		return -1;
+	if (!value)
+		value = "";
 	if( (strcmp(option, XAR_OPT_TOCCKSUM) == 0) ) {
+		xar_signature_t sig;
+
+		/* Cannot change XAR_OPT_TOCCKSUM after adding any files */
+		if (XAR(x)->files != NULL) {
+			xar_err_new(x);
+			xar_err_set_string(x, "XAR_OPT_TOCCKSUM must be set before files are added");
+			xar_err_callback(x, XAR_SEVERITY_WARNING, XAR_ERR_ARCHIVE_CREATION);
+			return -1;
+		}
+
 		if( strcmp(value, XAR_OPT_VAL_NONE) == 0 ) {
 			XAR(x)->heap_offset = 0;
+			XAR(x)->heap_len = 0;
 		}
-		if( strcmp(value, XAR_OPT_VAL_SHA1) == 0 ) {
-			XAR(x)->heap_offset = 20;
+		else {
+			const EVP_MD *md;
+			md = EVP_get_digestbyname(value);
+			if( md == NULL || EVP_MD_size(md) > HASH_MAX_MD_SIZE ) return -1;
+			XAR(x)->heap_offset = EVP_MD_size(md);
+			XAR(x)->heap_len = EVP_MD_size(md);
 		}
-		if( strcmp(value, XAR_OPT_VAL_MD5) == 0 ) {
-			XAR(x)->heap_offset = 16;
+
+		/* If there are any signatures already added they must have their offsets reset now */
+		sig = XAR(x)->signatures;
+		while (sig) {
+			XAR_SIGNATURE(sig)->offset = XAR(x)->heap_offset;
+			XAR(x)->heap_offset += XAR_SIGNATURE(sig)->len;
+			XAR(x)->heap_len += XAR_SIGNATURE(sig)->len;
+			sig = XAR_SIGNATURE(sig)->next;
 		}
 	}
-	for(i = XAR(x)->attrs; i ; i = XAR_ATTR(i)->next) {
-		if(strcmp(XAR_ATTR(i)->key, option)==0) {
-			free((char*)XAR_ATTR(i)->value);
-			XAR_ATTR(i)->value = strdup(value);
-			return 0;
+	if( (strcmp(option, XAR_OPT_FILECKSUM) == 0) ) {
+		if( strcmp(value, XAR_OPT_VAL_NONE) != 0 ) {
+			const EVP_MD *md;
+			md = EVP_get_digestbyname(value);
+			if( md == NULL || EVP_MD_size(md) > HASH_MAX_MD_SIZE ) return -1;
 		}
 	}
+	if ((strcmp(option, XAR_OPT_STRIPCOMPONENTS) == 0)) {
+		long comps;
+		char *endptr;
+		comps = strtol(value, &endptr, 0);
+		if (!*value || *endptr || comps < 0)
+			return -1;
+		XAR(x)->stripcomps = (int)comps;
+	}
+	if ((strcmp(option, XAR_OPT_EXTRACTSTDOUT) == 0)) {
+		XAR(x)->tostdout = strcmp(value, XAR_OPT_VAL_TRUE) == 0;
+	}
+	if ((XAR(x)->files == NULL && strcmp(option, XAR_OPT_RFC6713FORMAT) == 0)) {
+		XAR(x)->rfcformat = strcmp(value, XAR_OPT_VAL_TRUE) == 0;
+	}
 	a = xar_attr_new();
 	XAR_ATTR(a)->key = strdup(option);
 	XAR_ATTR(a)->value = strdup(value);
@@ -664,6 +920,26 @@ int32_t xar_opt_set(xar_t x, const char *option, const char *value) {
 	return 0;
 }
 
+/* xar_opt_unset
+ * x: the archive to set the option of
+ * option: the name of the option to delete
+ * This will delete ALL instances of the option name
+ */
+int32_t xar_opt_unset(xar_t x, const char *option) {
+	xar_attr_t i, p = NULL;
+	for(i = XAR(x)->attrs; i ; p = i, i = XAR_ATTR(i)->next) {
+		if(strcmp(XAR_ATTR(i)->key, option)==0) {
+			if( p == NULL )
+				XAR(x)->attrs = XAR_ATTR(i)->next;
+			else
+				XAR_ATTR(p)->next = XAR_ATTR(i)->next;
+			xar_attr_free(i);
+			i = p;
+		}
+	}
+	return 0;
+}
+
 /* xar_add_node
  * x: archive the file should belong to
  * f: parent node, possibly NULL
@@ -681,13 +957,15 @@ static xar_file_t xar_add_node(xar_t x, xar_file_t f, const char *name, const ch
 	char *tmp;
 	char idstr[32];
 
+	(void)srcpath;
 	if( !f ) {
+		int err;
 		if( realpath )
-			asprintf(&tmp, "%s", realpath);
+			err = asprintf(&tmp, "%s", realpath);
 		else
-			asprintf(&tmp, "%s%s%s", XAR(x)->path_prefix, prefix, name);
+			err = asprintf(&tmp, "%s%s%s", XAR(x)->path_prefix, prefix, name);
 
-		if( lstat(tmp, &XAR(x)->sbcache) != 0 ) {
+		if( err == -1 || lstat(tmp, &XAR(x)->sbcache) != 0 ) {
 			free(tmp);
 			return NULL;
 		}
@@ -700,16 +978,19 @@ static xar_file_t xar_add_node(xar_t x, xar_file_t f, const char *name, const ch
 		xar_attr_set(ret, NULL, "id", idstr);
 		XAR_FILE(ret)->parent = NULL;
 		XAR_FILE(ret)->fspath = tmp;
-		if( XAR(x)->files == NULL )
+		if( XAR(x)->files == NULL ) {
+			xar_check_force_rfc6713(x);
 			XAR(x)->files = ret;
+		}
 		else {
 			XAR_FILE(ret)->next = XAR(x)->files;
 			XAR(x)->files = ret;
 		}
 	} else {
+		int err;
 		path = XAR_FILE(f)->fspath;
 		if( strcmp(prefix, "../") == 0 ) {
-			int len1, len2;
+			size_t len1, len2;
 			len1 = strlen(path);
 			len2 = strlen(name);
 			if( (len1>=len2) && (strcmp(path+(len1-len2), name) == 0) ) {
@@ -719,11 +1000,11 @@ static xar_file_t xar_add_node(xar_t x, xar_file_t f, const char *name, const ch
 		}
 
 		if( realpath ){
-			asprintf(&tmp, "%s", realpath);
+			err = asprintf(&tmp, "%s", realpath);
 		}else
-			asprintf(&tmp, "%s/%s%s", path, prefix, name);
+			err = asprintf(&tmp, "%s/%s%s", path, prefix, name);
 		
-		if( lstat(tmp, &XAR(x)->sbcache) != 0 ) {
+		if( err == -1 || lstat(tmp, &XAR(x)->sbcache) != 0 ) {
 			free(tmp);
 			return NULL;
 		}
@@ -773,12 +1054,13 @@ static xar_file_t xar_add_pseudodir(xar_t x, xar_file_t f, const char *name, con
 	char idstr[32];
 
 	if( !f ) {
+		int err;
 		if( realpath )
-			asprintf(&tmp, "%s", realpath);
+			err = asprintf(&tmp, "%s", realpath);
 		else
-			asprintf(&tmp, "%s%s%s", XAR(x)->path_prefix, prefix, name);
+			err = asprintf(&tmp, "%s%s%s", XAR(x)->path_prefix, prefix, name);
 
-		if( lstat(tmp, &XAR(x)->sbcache) != 0 ) {
+		if( err == -1 || lstat(tmp, &XAR(x)->sbcache) != 0 ) {
 			free(tmp);
 			return NULL;
 		}
@@ -791,16 +1073,19 @@ static xar_file_t xar_add_pseudodir(xar_t x, xar_file_t f, const char *name, con
 		xar_attr_set(ret, NULL, "id", idstr);
 		XAR_FILE(ret)->parent = NULL;
 		XAR_FILE(ret)->fspath = tmp;
-		if( XAR(x)->files == NULL )
+		if( XAR(x)->files == NULL ) {
+			xar_check_force_rfc6713(x);
 			XAR(x)->files = ret;
+		}
 		else {
 			XAR_FILE(ret)->next = XAR(x)->files;
 			XAR(x)->files = ret;
 		}
 	} else {
+		int err;
 		path = XAR_FILE(f)->fspath;
 		if( strcmp(prefix, "../") == 0 ) {
-			int len1, len2;
+			size_t len1, len2;
 			len1 = strlen(path);
 			len2 = strlen(name);
 			if( (len1>=len2) && (strcmp(path+(len1-len2), name) == 0) ) {
@@ -810,11 +1095,11 @@ static xar_file_t xar_add_pseudodir(xar_t x, xar_file_t f, const char *name, con
 		}
 
 		if( realpath ){
-			asprintf(&tmp, "%s", realpath);
+			err = asprintf(&tmp, "%s", realpath);
 		}else
-			asprintf(&tmp, "%s/%s%s", path, prefix, name);
+			err = asprintf(&tmp, "%s/%s%s", path, prefix, name);
 		
-		if( lstat(tmp, &XAR(x)->sbcache) != 0 ) {
+		if( err == -1 || lstat(tmp, &XAR(x)->sbcache) != 0 ) {
 			free(tmp);
 			return NULL;
 		}
@@ -859,14 +1144,18 @@ static xar_file_t xar_add_r(xar_t x, xar_file_t f, const char *path, const char
 
 	if( strcmp(tmp3, "..") == 0 ) {
 		char *prefixstr;
+		int err;
 		if( !XAR(x)->skipwarn ) {
 			xar_err_new(x);
 			xar_err_set_string(x, "Skipping .. in path");
 			xar_err_callback(x, XAR_SEVERITY_WARNING, XAR_ERR_ARCHIVE_CREATION);
 			XAR(x)->skipwarn = 1;
 		}
-		asprintf(&prefixstr, "%s../", prefix);
-		ret2 = xar_add_r(x, f, tmp2, prefixstr);
+		err = asprintf(&prefixstr, "%s../", prefix);
+		if (err == -1)
+			ret2 = NULL;
+		else
+			ret2 = xar_add_r(x, f, tmp2, prefixstr);
 		free(prefixstr);
 		free(tmp1);
 		return ret2;
@@ -905,7 +1194,7 @@ static xar_file_t xar_add_r(xar_t x, xar_file_t f, const char *path, const char
 
 	/* tmp3 was not found in children of start, so we add it */
 	if( tmp2 ) {
-		//ret = xar_add_node(x, f, tmp3, prefix, NULL,  1);
+		/*ret = xar_add_node(x, f, tmp3, prefix, NULL,  1);*/
 		ret = xar_add_pseudodir(x, f, tmp3, prefix, NULL);
 	} else {
 		ret = xar_add_node(x, f, tmp3, prefix, NULL,  0);
@@ -979,8 +1268,10 @@ xar_file_t xar_add_frombuffer(xar_t x, xar_file_t parent, const char *name, char
 		snprintf(idstr, sizeof(idstr)-1, "%"PRIu64, ++XAR(x)->last_fileid);
 		xar_attr_set(ret, NULL, "id", idstr);
 		XAR_FILE(ret)->parent = NULL;
-		if( XAR(x)->files == NULL )
+		if( XAR(x)->files == NULL ) {
+			xar_check_force_rfc6713(x);
 			XAR(x)->files = ret;
+		}
 		else {
 			XAR_FILE(ret)->next = XAR(x)->files;
 			XAR(x)->files = ret;
@@ -997,7 +1288,7 @@ xar_file_t xar_add_frombuffer(xar_t x, xar_file_t parent, const char *name, char
 	
 	xar_prop_set(ret, "name", name);
 		
-	//int32_t xar_arcmod_archive(xar_t x, xar_file_t f, const char *file, const char *buffer, size_t len) 
+	/*int32_t xar_arcmod_archive(xar_t x, xar_file_t f, const char *file, const char *buffer, size_t len) */
 	if( xar_arcmod_archive(x, ret, NULL , buffer , length) < 0 ) {
 		xar_file_t i;
 		if( parent ) {
@@ -1034,8 +1325,10 @@ xar_file_t xar_add_folder(xar_t x, xar_file_t f, const char *name, struct stat *
 	if( !f ) {
 		XAR_FILE(ret)->parent = NULL;
 		
-		if( XAR(x)->files == NULL )
+		if( XAR(x)->files == NULL ) {
+			xar_check_force_rfc6713(x);
 			XAR(x)->files = ret;
+		}
 		else {
 			XAR_FILE(ret)->next = XAR(x)->files;
 			XAR(x)->files = ret;
@@ -1069,6 +1362,7 @@ xar_file_t xar_add_from_archive(xar_t x, xar_file_t parent, const char *name, xa
 {
 	xar_file_t ret;
 	char idstr[32];
+	xar_prop_t p;
 		
 	ret = xar_file_replicate(sourcefile, parent);
 	
@@ -1083,8 +1377,10 @@ xar_file_t xar_add_from_archive(xar_t x, xar_file_t parent, const char *name, xa
 	if( !parent ) {
 		XAR_FILE(ret)->parent = NULL;
 		
-		if( XAR(x)->files == NULL )
+		if( XAR(x)->files == NULL ) {
+			xar_check_force_rfc6713(x);
 			XAR(x)->files = ret;
+		}
 		else {
 			XAR_FILE(ret)->next = XAR(x)->files;
 			XAR(x)->files = ret;
@@ -1094,7 +1390,7 @@ xar_file_t xar_add_from_archive(xar_t x, xar_file_t parent, const char *name, xa
 	xar_prop_set(ret, "name", name);
 		
 	/* iterate through all the properties, see if any of them have an offset */
-	xar_prop_t p = xar_prop_pfirst(ret);
+	p = xar_prop_pfirst(ret);
 
 	do{
 		xar_prop_t tmpp;
@@ -1153,8 +1449,18 @@ int32_t xar_extract_tobuffer(xar_t x, xar_file_t f, char **buffer) {
 */
 int32_t xar_extract_tobuffersz(xar_t x, xar_file_t f, char **buffer, size_t *size) {
 	const char *sizestring = NULL;
-	
+	int32_t ret;
+
+	if (!buffer || !size)
+		return -1;
+
 	if(0 != xar_prop_get(f,"data/size",&sizestring)){
+		if(0 != xar_prop_get(f, "type", &sizestring))
+			return -1;
+		if(strcmp(sizestring, "file") == 0) {
+			*size = 0;
+			return 0;
+		}
 		return -1;
 	}
 
@@ -1164,8 +1470,56 @@ int32_t xar_extract_tobuffersz(xar_t x, xar_file_t f, char **buffer, size_t *siz
 	if(!(*buffer)){
 		return -1;
 	}
-	
-	return xar_arcmod_extract(x,f,NULL,*buffer,*size);
+	if (!*size)
+		return 0;
+
+	ret = xar_arcmod_extract(x,f,NULL,*buffer,*size);
+	if( ret ) {
+		*size = 0;
+		free(*buffer);
+		*buffer = NULL;
+	}
+
+	return ret;
+}
+
+int32_t xar_extract_tostream_init(xar_t x, xar_file_t f, xar_stream *stream) {
+	xar_prop_t tmpp;
+
+	if( !xar_check_prop(x, "data") )
+		return XAR_STREAM_OK;
+
+	tmpp = xar_prop_pfirst(f);
+	if( tmpp )
+		tmpp = xar_prop_find(tmpp, "data");
+	if( !tmpp )
+		return XAR_STREAM_OK;
+
+	return xar_attrcopy_from_heap_to_stream_init(x, f, tmpp, stream);
+}
+
+int32_t xar_extract_tostream(xar_stream *stream) {
+	return xar_attrcopy_from_heap_to_stream(stream);
+}
+
+int32_t xar_extract_tostream_end(xar_stream *stream) {
+	return xar_attrcopy_from_heap_to_stream_end(stream);
+}
+
+const char *xar_strip_components(const char *path, int components)
+{
+	while (components-- > 0) {
+		const char *nextcomp;
+		nextcomp = strchr(path, '/');
+		if (!nextcomp)
+			return NULL;
+		while (nextcomp[1] == '/')
+			++nextcomp;
+		if (!nextcomp[1])
+			return NULL;
+		path = nextcomp + 1;
+	}
+	return path;
 }
 
 /* xar_extract
@@ -1184,22 +1538,39 @@ int32_t xar_extract(xar_t x, xar_file_t f) {
 	struct stat sb;
 	char *tmp1, *dname;
 	xar_file_t tmpf;
-	
-	if( (strstr(XAR_FILE(f)->fspath, "/") != NULL) && (stat(XAR_FILE(f)->fspath, &sb)) && (XAR_FILE(f)->parent_extracted == 0) ) {
-		tmp1 = strdup(XAR_FILE(f)->fspath);
-		dname = dirname(tmp1);
-		tmpf = xar_file_find(XAR(x)->files, dname);
-		if( !tmpf ) {
-			xar_err_set_string(x, "Unable to find file");
-			xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
-			return -1;
+	const char *fspath;
+
+	fspath = XAR_FILE(f)->fspath;
+	if (XAR(x)->stripcomps && (fspath=xar_strip_components(fspath, XAR(x)->stripcomps)) == NULL)
+		return -1;
+	if (XAR(x)->tostdout) {
+		char *buffer = NULL;
+		size_t bsize = 0;
+		int32_t result;
+		if ((result = xar_extract_tobuffersz(x, f, &buffer, &bsize)) != 0)
+			return result;
+		if (bsize && buffer && fwrite(buffer, bsize, 1, stdout) != 1)
+			result = -1;
+		free(buffer);
+		return result;
+	}
+	else {
+		if( (strstr(fspath, "/") != NULL) && (stat(fspath, &sb)) && (XAR_FILE(f)->parent_extracted == 0) ) {
+			tmp1 = strdup(XAR_FILE(f)->fspath);
+			dname = dirname(tmp1);
+			tmpf = xar_file_find(XAR(x)->files, dname);
+			if( !tmpf ) {
+				xar_err_set_string(x, "Unable to find file");
+				xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
+				return -1;
+			}
+			free(tmp1);
+			XAR_FILE(f)->parent_extracted++;
+			xar_extract(x, tmpf);
 		}
-		free(tmp1);
-		XAR_FILE(f)->parent_extracted++;
-		xar_extract(x, tmpf);
+
+		return xar_extract_tofile(x, f, fspath);
 	}
-	
-	return xar_extract_tofile(x, f, XAR_FILE(f)->fspath);
 }
 
 /* xar_verify
@@ -1228,23 +1599,23 @@ static int toc_read_callback(void *context, char *buffer, int len) {
 	if ( ((!XAR(x)->offset) || (XAR(x)->offset == XAR(x)->readbuf_len)) && (XAR(x)->toc_count != XAR(x)->header.toc_length_compressed) ) {
 		XAR(x)->offset = 0;
 		if( (XAR(x)->readbuf_len - off) + XAR(x)->toc_count > XAR(x)->header.toc_length_compressed )
-			ret = xar_read_fd(XAR(x)->fd, XAR(x)->readbuf, XAR(x)->header.toc_length_compressed - XAR(x)->toc_count);
+			ret = (int)xar_read_fd(XAR(x)->fd, XAR(x)->readbuf, (size_t)(XAR(x)->header.toc_length_compressed - XAR(x)->toc_count));
 		else
-			ret = read(XAR(x)->fd, XAR(x)->readbuf, XAR(x)->readbuf_len);
+			ret = (int)read(XAR(x)->fd, XAR(x)->readbuf, XAR(x)->readbuf_len);
 		if ( ret == -1 )
 			return ret;
 
 		if ( XAR(x)->docksum )
-			EVP_DigestUpdate(&XAR(x)->toc_ctx, XAR(x)->readbuf, ret);
+			EVP_DigestUpdate(XAR(x)->toc_ctx, XAR(x)->readbuf, ret);
 
 		XAR(x)->toc_count += ret;
 		off += ret;
 	}
 
-	if( off && (off < XAR(x)->readbuf_len) )
+	if( off && (off < (int)XAR(x)->readbuf_len) )
 		XAR(x)->readbuf_len = off;
-	XAR(x)->zs.next_in = XAR(x)->readbuf + XAR(x)->offset;
-	XAR(x)->zs.avail_in = XAR(x)->readbuf_len - XAR(x)->offset;
+	XAR(x)->zs.next_in = ((unsigned char *)XAR(x)->readbuf) + XAR(x)->offset;
+	XAR(x)->zs.avail_in = (unsigned)(XAR(x)->readbuf_len - XAR(x)->offset);
 	XAR(x)->zs.next_out = (void *)buffer;
 	XAR(x)->zs.avail_out = len;
 
@@ -1263,6 +1634,7 @@ static int toc_read_callback(void *context, char *buffer, int len) {
  * Summary: this is the callback for xmlTextReaderForIO to close the IO
  */
 static int close_callback(void *context) {
+	(void)context;
 	return 0;
 }
 
@@ -1367,6 +1739,7 @@ static int32_t xar_unserialize(xar_t x) {
 				} else {
 					xar_subdoc_t s;
 					int i;
+					struct __xar_attr_t *attrs = NULL;
 
 					prefix = xmlTextReaderPrefix(reader);
 					uri = xmlTextReaderNamespaceUri(reader);
@@ -1384,13 +1757,14 @@ static int32_t xar_unserialize(xar_t x) {
 								a = xar_attr_new();
 								XAR_ATTR(a)->key = strdup(aname);
 								XAR_ATTR(a)->value = strdup(avalue);
-								XAR_ATTR(a)->next = XAR_SUBDOC(s)->attrs;
-								XAR_SUBDOC(s)->attrs = XAR_ATTR(a);
+								XAR_ATTR(a)->next = attrs;
+								attrs = XAR_ATTR(a);
 							}
 						}
 					}
 
 					s = xar_subdoc_new(x, (const char *)name);
+					XAR_SUBDOC(s)->attrs = attrs;
 					xar_subdoc_unserialize(s, reader);
 				}
 			}
