@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005 Rob Braun
+ * Copyright (c) 2005-2008 Rob Braun
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -28,7 +28,7 @@
  */
 /*
  * 03-Apr-2005
- * DRI: Rob Braun <bbraun@opendarwin.org>
+ * DRI: Rob Braun <bbraun@synack.net>
  */
 /*
  * Portions Copyright 2006, Apple Computer, Inc.
@@ -37,6 +37,7 @@
 
 #define _FILE_OFFSET_BITS 64
 
+#include "config.h"
 #include <stdlib.h>
 #include <stdio.h>
 #include <fcntl.h>
@@ -48,7 +49,6 @@
 #include <sys/types.h>
 #include <assert.h>
 
-#include "config.h"
 #ifndef HAVE_ASPRINTF
 #include "asprintf.h"
 #endif
@@ -58,9 +58,11 @@
 #include "io.h"
 #include "zxar.h"
 #include "bzxar.h"
+#include "lzmaxar.h"
 #include "hash.h"
 #include "script.h"
 #include "macho.h"
+#include "util.h"
 
 #if !defined(LLONG_MAX) && defined(LONG_LONG_MAX)
 #define LLONG_MAX LONG_LONG_MAX
@@ -70,7 +72,7 @@
 #define LLONG_MIN LONG_LONG_MIN
 #endif
 
-struct datamod xar_datamods[] = {
+static const struct datamod xar_datamods[] = {
 	{ xar_hash_archived,
 	  xar_hash_unarchived_out,
 	  xar_hash_out_done,
@@ -105,33 +107,138 @@ struct datamod xar_datamods[] = {
 	  xar_bzip_toheap_in,
 	  (toheap_out)NULL,
 	  xar_bzip_toheap_done
+	},
+	{ xar_lzma_fromheap_in,
+	  (fromheap_out)NULL,
+	  xar_lzma_fromheap_done,
+	  xar_lzma_toheap_in,
+	  (toheap_out)NULL,
+	  xar_lzma_toheap_done
 	}
 };
 
+static const is_compressed xar_compresschecks[] = {
+	xar_gzip_is_compressed,
+	xar_bzip_is_compressed,
+	xar_lzma_is_compressed
+};
+
+
+static size_t get_rsize(xar_t x) {
+	size_t bsize;
+	const char *opt = NULL;
+
+	opt = xar_opt_get(x, "rsize");
+	if( !opt ) {
+		bsize = XAR_DEFAULT_BUFFER_SIZE;
+	} else {
+		bsize = strtol(opt, NULL, 0);
+		if( ((bsize == (size_t)LONG_MAX) || (bsize == (size_t)LONG_MIN)) && (errno == ERANGE) ) {
+			bsize = XAR_DEFAULT_BUFFER_SIZE;
+		}
+		if (bsize < XAR_MINIMUM_BUFFER_SIZE)
+			bsize = XAR_MINIMUM_BUFFER_SIZE;
+	}
+
+	return bsize;
+}
+	
+static off_t get_offset(xar_t x, xar_file_t f, xar_prop_t p) {
+	off_t seekoff;
+	xar_prop_t tmpp;
+	const char *opt = NULL;
+
+	(void)x; (void)f;
+	tmpp = xar_prop_pget(p, "offset");
+	if( tmpp )
+		opt = xar_prop_getvalue(tmpp);
+	
+	seekoff = strtoll(opt, NULL, 0);
+		
+	if( ((seekoff == LLONG_MAX) || (seekoff == LLONG_MIN)) && (errno == ERANGE) ) {
+		return -1;
+	}
+
+	return seekoff;
+}
+
+static int64_t get_length(xar_prop_t p) {
+	const char *opt = NULL;
+	int64_t fsize = 0;
+	xar_prop_t tmpp;
+
+	tmpp = xar_prop_pget(p, "length");
+	if( tmpp )
+		opt = xar_prop_getvalue(tmpp);
+	if( !opt ) {
+		return 0;
+	} else {
+		fsize = strtoll(opt, NULL, 10);
+		if( ((fsize == LLONG_MAX) || (fsize == LLONG_MIN)) && (errno == ERANGE) ) {
+			return -1;
+		}
+	}
+
+	return fsize;
+}
+
+static void xar_io_seek(xar_t x, xar_file_t f, off_t seekoff) {
+	int r;
+
+	if( XAR(x)->fd >= 0 ) {
+		r = (int)lseek(XAR(x)->fd, seekoff, SEEK_SET);
+		if( r == -1 ) {
+			if( errno == ESPIPE ) {
+				ssize_t rr;
+				char *buf;
+				unsigned int len;
+
+				len = (unsigned)(seekoff - (off_t)xar_get_heap_offset(x));
+				if( XAR(x)->heap_offset > len ) {
+					xar_err_new(x);
+					xar_err_set_file(x, f);
+					xar_err_set_string(x, "Unable to seek");
+					xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
+				} else {
+					len -= (unsigned)XAR(x)->heap_offset;
+					buf = malloc(len);
+					assert(buf);
+					rr = xar_read_fd(XAR(x)->fd, buf, len);
+					if( rr < (ssize_t)len ) {
+						xar_err_new(x);
+						xar_err_set_file(x, f);
+						xar_err_set_string(x, "Unable to seek");
+						xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
+					}
+					XAR(x)->heap_offset += rr;
+					free(buf);
+				}
+			} else {
+				xar_err_new(x);
+				xar_err_set_file(x, f);
+				xar_err_set_string(x, "Unable to seek");
+				xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
+			}
+		}
+	}
+}
+
 int32_t xar_attrcopy_to_heap(xar_t x, xar_file_t f, xar_prop_t p, read_callback rcb, void *context) {
-	int modulecount = (sizeof(xar_datamods)/sizeof(struct datamod));
-	void	*modulecontext[modulecount];
+	void	*modulecontext[sizeof(xar_datamods)/sizeof(struct datamod)];
+	int modulecount = (int)(sizeof(modulecontext)/sizeof(modulecontext[0]));
 	int r, off, i;
 	size_t bsize, rsize;
 	int64_t readsize=0, writesize=0, inc = 0;
 	void *inbuf;
 	char *tmpstr = NULL;
-	const char *opt, *csum;
+	const char *opt = NULL, *csum = NULL;
 	off_t orig_heap_offset = XAR(x)->heap_offset;
 	xar_file_t tmpf = NULL;
 	xar_prop_t tmpp = NULL;
 
 	memset(modulecontext, 0, sizeof(void*)*modulecount);
 
-	opt = xar_opt_get(x, XAR_OPT_RSIZE);
-	if( !opt ) {
-		bsize = 4096;
-	} else {
-		bsize = strtol(opt, NULL, 0);
-		if( ((bsize == LONG_MAX) || (bsize == LONG_MIN)) && (errno == ERANGE) ) {
-			bsize = 4096;
-		}
-	}
+	bsize = get_rsize(x);
 
 	r = 1;
 	while(r != 0) {
@@ -165,12 +272,12 @@ int32_t xar_attrcopy_to_heap(xar_t x, xar_file_t f, xar_prop_t p, read_callback
 		off = 0;
 		if( rsize != 0 ) {
 			do {
-				r = write(XAR(x)->heap_fd, inbuf+off, rsize-off);
+				r = (int)write(XAR(x)->heap_fd, ((char *)inbuf)+off, rsize-off);
 				if( (r < 0) && (errno != EINTR) )
 					return -1;
 				off += r;
 				writesize += r;
-			} while( off < rsize );
+			} while( (size_t)off < rsize );
 		}
 		XAR(x)->heap_offset += off;
 		free(inbuf);
@@ -198,7 +305,8 @@ int32_t xar_attrcopy_to_heap(xar_t x, xar_file_t f, xar_prop_t p, read_callback
 	tmpp = xar_prop_pget(p, "archived-checksum");
 	if( tmpp )
 		csum = xar_prop_getvalue(tmpp);
-	tmpf = xmlHashLookup(XAR(x)->csum_hash, BAD_CAST(csum));
+	if( csum )
+		tmpf = xmlHashLookup(XAR(x)->csum_hash, BAD_CAST(csum));
 	if( tmpf ) {
 		const char *attr = xar_prop_getkey(p);
 		opt = xar_opt_get(x, XAR_OPT_LINKSAME);
@@ -251,27 +359,37 @@ int32_t xar_attrcopy_to_heap(xar_t x, xar_file_t f, xar_prop_t p, read_callback
 			}
 			
 		}
-	} else {
+	} else if( csum ) {
 		xmlHashAddEntry(XAR(x)->csum_hash, BAD_CAST(csum), XAR_FILE(f));
+	} else {
+		xar_err_new(x);
+		xar_err_set_file(x, f);
+		xar_err_set_string(x, "No archived-checksum");
+		xar_err_callback(x, XAR_SEVERITY_WARNING, XAR_ERR_ARCHIVE_CREATION);
 	}
 
-	asprintf(&tmpstr, "%"PRIu64, readsize);
+	if (asprintf(&tmpstr, "%"PRIu64, readsize) == -1)
+		return -1;
 	xar_prop_pset(f, p, "size", tmpstr);
 	free(tmpstr);
 
-	asprintf(&tmpstr, "%"PRIu64, (uint64_t)orig_heap_offset);
+	if (asprintf(&tmpstr, "%"PRIu64, (uint64_t)orig_heap_offset) == -1)
+		return -1;
 	xar_prop_pset(f, p, "offset", tmpstr);
 	free(tmpstr);
-	
-	tmpstr = (char *)xar_opt_get(x, XAR_OPT_COMPRESSION);
-	if( tmpstr && (strcmp(tmpstr, XAR_OPT_VAL_NONE) == 0) ) {
+
+	tmpp = xar_prop_pget(p, "encoding");
+	if (!tmpp) {
 		xar_prop_pset(f, p, "encoding", NULL);
 		tmpp = xar_prop_pget(p, "encoding");
-		if( tmpp )
-			xar_attr_pset(f, tmpp, "style", "application/octet-stream");
+	}
+	opt = xar_attr_pget(f, tmpp, "style");
+	if (tmpp && (!opt || !*opt)) {
+		xar_attr_pset(f, tmpp, "style", "application/octet-stream");
 	}
 
-	asprintf(&tmpstr, "%"PRIu64, writesize);
+	if (asprintf(&tmpstr, "%"PRIu64, writesize) == -1)
+		return -1;
 	xar_prop_pset(f, p, "length", tmpstr);
 	free(tmpstr);
 
@@ -284,8 +402,8 @@ int32_t xar_attrcopy_to_heap(xar_t x, xar_file_t f, xar_prop_t p, read_callback
  * It is assumed the heap_fd is already positioned appropriately.
  */
 int32_t xar_attrcopy_from_heap(xar_t x, xar_file_t f, xar_prop_t p, write_callback wcb, void *context) {
-	int modulecount = (sizeof(xar_datamods)/sizeof(struct datamod));
-	void	*modulecontext[modulecount];
+	void	*modulecontext[sizeof(xar_datamods)/sizeof(struct datamod)];
+	int modulecount = (int)(sizeof(modulecontext)/sizeof(modulecontext[0]));
 	int r, i;
 	size_t bsize, def_bsize;
 	int64_t fsize, inc = 0, seekoff;
@@ -295,15 +413,7 @@ int32_t xar_attrcopy_from_heap(xar_t x, xar_file_t f, xar_prop_t p, write_callba
 
 	memset(modulecontext, 0, sizeof(void*)*modulecount);
 
-	opt = xar_opt_get(x, "rsize");
-	if( !opt ) {
-		def_bsize = 4096;
-	} else {
-		def_bsize = strtol(opt, NULL, 0);
-		if( ((def_bsize == LONG_MAX) || (def_bsize == LONG_MIN)) && (errno == ERANGE) ) {
-			def_bsize = 4096;
-		}
-	}
+	def_bsize = get_rsize(x);
 
 	opt = NULL;
 	tmpp = xar_prop_pget(p, "offset");
@@ -319,56 +429,14 @@ int32_t xar_attrcopy_from_heap(xar_t x, xar_file_t f, xar_prop_t p, write_callba
 		}
 	}
 
-	seekoff += XAR(x)->toc_count + sizeof(xar_header_t);
-	if( XAR(x)->fd > 1 ) {
-		r = lseek(XAR(x)->fd, seekoff, SEEK_SET);
-		if( r == -1 ) {
-			if( errno == ESPIPE ) {
-				ssize_t rr;
-				char *buf;
-				unsigned int len;
+	seekoff += (int64_t)xar_get_heap_offset(x);
+	xar_io_seek(x, f, seekoff);
 
-				len = seekoff - XAR(x)->toc_count;
-				len -= sizeof(xar_header_t);
-				if( XAR(x)->heap_offset > len ) {
-					xar_err_new(x);
-					xar_err_set_file(x, f);
-					xar_err_set_string(x, "Unable to seek");
-					xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
-				} else {
-					len -= XAR(x)->heap_offset;
-					buf = malloc(len);
-					assert(buf);
-					rr = read(XAR(x)->fd, buf, len);
-					if( rr < len ) {
-						xar_err_new(x);
-						xar_err_set_file(x, f);
-						xar_err_set_string(x, "Unable to seek");
-						xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
-					}
-					free(buf);
-				}
-			} else {
-				xar_err_new(x);
-				xar_err_set_file(x, f);
-				xar_err_set_string(x, "Unable to seek");
-				xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
-			}
-		}
-	}
-
-	opt = NULL;
-	tmpp = xar_prop_pget(p, "length");
-	if( tmpp )
-		opt = xar_prop_getvalue(tmpp);
-	if( !opt ) {
+	fsize = get_length(p);
+	if( fsize == 0 )
 		return 0;
-	} else {
-		fsize = strtoll(opt, NULL, 10);
-		if( ((fsize == LLONG_MAX) || (fsize == LLONG_MIN)) && (errno == ERANGE) ) {
-			return -1;
-		}
-	}
+	if( fsize < 0 )
+		return -1;
 
 	bsize = def_bsize;
 	inbuf = malloc(bsize);
@@ -380,9 +448,9 @@ int32_t xar_attrcopy_from_heap(xar_t x, xar_file_t f, xar_prop_t p, write_callba
 		/* Size has been reached */
 		if( fsize == inc )
 			break;
-		if( (fsize - inc) < bsize )
-			bsize = fsize - inc;
-		r = read(XAR(x)->fd, inbuf, bsize);
+		if( (fsize - inc) < (int64_t)bsize )
+			bsize = (size_t)(fsize - inc);
+		r = (int)read(XAR(x)->fd, inbuf, bsize);
 		if( r == 0 )
 			break;
 		if( (r < 0) && (errno == EINTR) )
@@ -454,77 +522,20 @@ int32_t xar_attrcopy_from_heap_to_heap(xar_t xsource, xar_file_t fsource, xar_pr
 	char *tmpstr = NULL;
 	xar_prop_t tmpp;
 	
-	opt = xar_opt_get(xsource, "rsize");
-	if( !opt ) {
-		bsize = 4096;
-	} else {
-		bsize = strtol(opt, NULL, 0);
-		if( ((bsize == LONG_MAX) || (bsize == LONG_MIN)) && (errno == ERANGE) ) {
-			bsize = 4096;
-		}
-	}
+	bsize = get_rsize(xsource);
 	
-	tmpp = xar_prop_pget(p, "offset");
-	if( tmpp )
-		opt = xar_prop_getvalue(tmpp);
-	
-	seekoff = strtoll(opt, NULL, 0);
-		
-	if( ((seekoff == LLONG_MAX) || (seekoff == LLONG_MIN)) && (errno == ERANGE) ) {
+	seekoff = get_offset(xsource, fsource, p);
+	if( seekoff < 0 )
 		return -1;
-	}
 	
-	seekoff += XAR(xsource)->toc_count + sizeof(xar_header_t);
+	seekoff += (int64_t)xar_get_heap_offset(xsource);
+	xar_io_seek(xsource, fsource, seekoff);
 	
-	if( XAR(xsource)->fd > 1 ) {
-		r = lseek(XAR(xsource)->fd, seekoff, SEEK_SET);
-		if( r == -1 ) {
-			if( errno == ESPIPE ) {
-				ssize_t rr;
-				char *buf;
-				unsigned int len;
-				
-				len = seekoff - XAR(xsource)->toc_count;
-				len -= sizeof(xar_header_t);
-				if( XAR(xsource)->heap_offset > len ) {
-					xar_err_new(xsource);
-					xar_err_set_file(xsource, fsource);
-					xar_err_set_string(xsource, "Unable to seek");
-					xar_err_callback(xsource, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
-				} else {
-					len -= XAR(xsource)->heap_offset;
-					buf = malloc(len);
-					assert(buf);
-					rr = read(XAR(xsource)->fd, buf, len);
-					if( rr < len ) {
-						xar_err_new(xsource);
-						xar_err_set_file(xsource, fsource);
-						xar_err_set_string(xsource, "Unable to seek");
-						xar_err_callback(xsource, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
-					}
-					free(buf);
-				}
-			} else {
-				xar_err_new(xsource);
-				xar_err_set_file(xsource, fsource);
-				xar_err_set_string(xsource, "Unable to seek");
-				xar_err_callback(xsource, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
-			}
-		}
-	}
-	
-	opt = NULL;
-	tmpp = xar_prop_pget(p, "length");
-	if( tmpp )
-		opt = xar_prop_getvalue(tmpp);
-	if( !opt ) {
+	fsize = get_length(p);
+	if( fsize == 0 )
 		return 0;
-	} else {
-		fsize = strtoll(opt, NULL, 10);
-		if( ((fsize == LLONG_MAX) || (fsize == LLONG_MIN)) && (errno == ERANGE) ) {
-			return -1;
-		}
-	}
+	if( fsize < 0 )
+		return -1;
 	
 	inbuf = malloc(bsize);
 	if( !inbuf ) {
@@ -536,9 +547,9 @@ int32_t xar_attrcopy_from_heap_to_heap(xar_t xsource, xar_file_t fsource, xar_pr
 		/* Size has been reached */
 		if( fsize == inc )
 			break;
-		if( (fsize - inc) < bsize )
-			bsize = fsize - inc;
-		r = read(XAR(xsource)->fd, inbuf, bsize);
+		if( (fsize - inc) < (int64_t)bsize )
+			bsize = (size_t)(fsize - inc);
+		r = (int)read(XAR(xsource)->fd, inbuf, bsize);
 		if( r == 0 )
 			break;
 		if( (r < 0) && (errno == EINTR) )
@@ -555,7 +566,7 @@ int32_t xar_attrcopy_from_heap_to_heap(xar_t xsource, xar_file_t fsource, xar_pr
 		off = 0;
 		
 		do {
-			r = write(XAR(xdest)->heap_fd, inbuf+off, r-off );
+			r = (int)write(XAR(xdest)->heap_fd, ((char *)inbuf)+off, r-off );
 			off += r;
 			writesize += r;
 		} while( off < r );
@@ -563,7 +574,10 @@ int32_t xar_attrcopy_from_heap_to_heap(xar_t xsource, xar_file_t fsource, xar_pr
 		XAR(xdest)->heap_len += off;
 	}
 	
-	asprintf(&tmpstr, "%"PRIu64, (uint64_t)orig_heap_offset);
+	if (asprintf(&tmpstr, "%"PRIu64, (uint64_t)orig_heap_offset) == -1) {
+		free(inbuf);
+		return -1;
+	}
 	opt = xar_prop_getkey(p);
 	tmpp = xar_prop_pfirst(fdest);
 	if( tmpp )
@@ -579,6 +593,199 @@ int32_t xar_attrcopy_from_heap_to_heap(xar_t xsource, xar_file_t fsource, xar_pr
 	
 	return 0;
 }
+
+static int32_t flush_stream(xar_stream *stream) {
+	xar_stream_state_t *state = (xar_stream_state_t *)(stream->state);
+
+	if( state->pending_buf && stream->avail_out ) {
+		size_t len = state->pending_buf_size;
+
+		if( stream->avail_out < len ) {
+			len = stream->avail_out;
+		}
+
+		memcpy(stream->next_out, state->pending_buf, len);
+		stream->next_out += len;
+		stream->avail_out -= (unsigned)len;
+		stream->total_out += len;
+
+		if( state->pending_buf_size == len )  {
+			state->pending_buf_size = 0;
+
+			free(state->pending_buf);
+			state->pending_buf = NULL;
+		} else if( state->pending_buf_size > len ) {
+			state->pending_buf_size -= len;
+			memcpy(state->pending_buf, state->pending_buf + len, state->pending_buf_size);
+		}
+	}
+
+	return XAR_STREAM_OK;
+}
+
+static int32_t write_to_stream(void *inbuf, size_t inlen, xar_stream *stream) {
+	xar_stream_state_t *state = (xar_stream_state_t *)stream->state;
+	size_t len = inlen;
+
+	if( stream->avail_out < len ) {
+		len = stream->avail_out;
+	}
+
+	memcpy(stream->next_out, inbuf, len);
+	stream->next_out += len;
+	stream->avail_out -= (unsigned)len;
+	stream->total_out += len;
+
+	if( inlen > len ) {
+		state->pending_buf_size = inlen - len;
+		state->pending_buf = malloc(state->pending_buf_size);
+
+		memcpy(state->pending_buf, (char *)inbuf + len, state->pending_buf_size);
+	}
+
+	return XAR_STREAM_OK;
+}
+
+int32_t xar_attrcopy_from_heap_to_stream_init(xar_t x, xar_file_t f, xar_prop_t p, xar_stream *stream) {
+	xar_stream_state_t *state;
+	off_t seekoff;
+
+	seekoff = get_offset(x, f, p);
+	if( seekoff < 0 ) 
+		return XAR_STREAM_ERR;
+
+	state = calloc(1, sizeof(xar_stream_state_t));
+	if( !state ) {
+		return XAR_STREAM_ERR;
+	}
+	stream->state = (void*)state;
+	state->bsize = get_rsize(x);
+
+	state->modulecount = (int)(sizeof(xar_datamods)/sizeof(struct datamod));
+	state->modulecontext = calloc(1, sizeof(void*)*state->modulecount);
+	if( !state->modulecontext ) {
+		free(state);
+		return XAR_STREAM_ERR;
+	}
+
+	seekoff += (off_t)xar_get_heap_offset(x);
+	xar_io_seek(x, f, seekoff);
+
+	stream->total_in = 0;
+	stream->total_out = 0;
+
+	state->fsize = get_length(p);
+
+	if(state->fsize == 0) {
+		return XAR_STREAM_OK;
+	} else if(state->fsize == -1) {
+		free(state->modulecontext);
+		free(state);
+		return XAR_STREAM_ERR;
+	}
+
+	state->pending_buf = NULL;
+	state->pending_buf_size = 0;
+	state->x = x;
+	state->f = f;
+	state->p = p;
+
+	return XAR_STREAM_OK;
+}
+
+int32_t xar_attrcopy_from_heap_to_stream(xar_stream *stream) {
+	xar_stream_state_t *state = stream->state;
+
+	int r, i;
+	size_t bsize;
+	void *inbuf; 
+
+	if( state->pending_buf_size ) {
+		return flush_stream(stream);
+	} 
+
+	bsize = state->bsize;
+	inbuf = malloc(bsize);
+	if( !inbuf ) {
+		return XAR_STREAM_ERR;
+	}
+        
+	/* Size has been reached */
+	if( (uint64_t)state->fsize == stream->total_in ) {
+		free(inbuf);
+		return XAR_STREAM_END;
+	}
+	if( (state->fsize - stream->total_in) < bsize )
+		bsize = (size_t)(state->fsize - stream->total_in);
+	r = (int)read(XAR(state->x)->fd, inbuf, bsize);
+	if( r == 0 ) {
+		free(inbuf);
+		return XAR_STREAM_END;
+	}
+	if( (r < 0) && (errno == EINTR) ) {
+		free(inbuf);
+		return XAR_STREAM_OK;
+	}
+	if( r < 0 ) {
+		free(inbuf);
+		return XAR_STREAM_ERR;
+	}
+
+	XAR(state->x)->heap_offset += r;
+	stream->total_in += r;
+	bsize = r;
+	
+	/* filter the data through the in modules */
+	for( i = 0; i < state->modulecount; i++) {
+	if( xar_datamods[i].fh_in ) {
+		int32_t ret;
+		ret = xar_datamods[i].fh_in(state->x, state->f, state->p, &inbuf, &bsize, &(state->modulecontext[i]));
+		if( ret < 0 )
+			return XAR_STREAM_ERR;
+		}
+	}
+
+	/* filter the data through the out modules */
+	for( i = 0; i < state->modulecount; i++) {
+		if( xar_datamods[i].fh_out ) {
+			int32_t ret;
+			ret = xar_datamods[i].fh_out(state->x, state->f, state->p, inbuf, bsize, &(state->modulecontext[i]));
+			if( ret < 0 )
+				return XAR_STREAM_ERR;
+		}
+	}
+
+	write_to_stream(inbuf, bsize, stream);
+
+	free(inbuf);
+
+	return XAR_STREAM_OK;
+}
+
+int32_t xar_attrcopy_from_heap_to_stream_end(xar_stream *stream) {
+	xar_stream_state_t *state = (xar_stream_state_t *)stream->state;
+	int i;
+
+	/* finish up anything that still needs doing */
+	for( i = 0; i < state->modulecount; i++) {
+		if( xar_datamods[i].fh_done ) {
+			int32_t ret;
+			ret = xar_datamods[i].fh_done(state->x, state->f, state->p, &(state->modulecontext[i]));
+			if( ret < 0 )
+				return ret;
+		}
+	}
+
+	if( state->pending_buf ) {
+		free(state->pending_buf);
+	}
+ 
+	free(state->modulecontext);
+	free(state);
+       
+	return XAR_STREAM_OK;
+}
+
 /* xar_heap_to_archive
  * x: archive to operate on
  * Returns 0 on success, -1 on error
@@ -593,12 +800,14 @@ int32_t xar_heap_to_archive(xar_t x) {
 
 	opt = xar_opt_get(x, "rsize");
 	if( !opt ) {
-		bsize = 4096;
+		bsize = XAR_DEFAULT_BUFFER_SIZE;
 	} else {
 		bsize = strtol(opt, NULL, 0);
 		if( ((bsize == LONG_MAX) || (bsize == LONG_MIN)) && (errno == ERANGE) ) {
-			bsize = 4096;
+			bsize = XAR_DEFAULT_BUFFER_SIZE;
 		}
+		if (bsize < XAR_MINIMUM_BUFFER_SIZE)
+			bsize = XAR_MINIMUM_BUFFER_SIZE;
 	}
 
 	b = malloc(bsize);
@@ -618,8 +827,36 @@ int32_t xar_heap_to_archive(xar_t x) {
 			r = write(XAR(x)->fd, b+off, bsize-off);
 			if( (r < 0) && (errno != EINTR) )
 				return -1;
-			off += r;
+			off += (int)r;
 		} while( off < bsize );
 	}
 	return 0;
 }
+
+/* xar_prevent_recompress
+ * x: archive to operate on
+ * Returns 0 to allow compression, non-zero to prevent compression
+ * Summary: prevents recompression unless XAR_OPT_RECOMPRESS is XAR_OPT_VAL_TRUE
+ */
+int32_t xar_prevent_recompress(xar_t x, void *in, size_t inlen) {
+	int checkcount = (int)(sizeof(xar_compresschecks)/sizeof(xar_compresschecks[0]));
+	int i;
+	const char *opt;
+	int recompress = 0;
+
+	opt = xar_opt_get(x, XAR_OPT_RECOMPRESS);
+	if( opt && (strcmp(opt, XAR_OPT_VAL_TRUE) == 0) ) {
+		recompress = 1;
+	}
+
+	if (recompress)
+		return 0;
+
+	/* check with the modules */
+	for( i = 0; i < checkcount; ++i) {
+		if ((*xar_compresschecks[i])(in, inlen))
+			return 1;
+	}
+
+	return 0;
+}
