@@ -28,7 +28,7 @@
  */
 /*
  * 03-Apr-2005
- * DRI: Rob Braun <bbraun@opendarwin.org>
+ * DRI: Rob Braun <bbraun@synack.net>
  */
 /*
  * Portions Copyright 2003, Apple Computer, Inc.
@@ -53,7 +53,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/time.h>
-#include <sys/fcntl.h>
+#include <fcntl.h>
 #include <time.h>
 #include <pwd.h>
 #include <grp.h>
@@ -71,6 +71,18 @@
 #include "archive.h"
 #include "util.h"
 
+#ifdef __APPLE__
+#ifndef UF_COMPRESSED
+#define UF_COMPRESSED 0x00000020
+#endif
+#ifndef UF_TRACKED
+#define UF_TRACKED 0x00000040
+#endif
+#ifndef UF_HIDDEN
+#define UF_HIDDEN 0x00008000
+#endif
+#endif
+
 #ifndef LLONG_MIN
 #define LLONG_MIN LONG_LONG_MIN
 #endif
@@ -118,7 +130,7 @@ static xar_file_t xar_link_lookup(xar_t x, dev_t dev, ino_t ino, xar_file_t f) {
 	return ret;
 }
 
-static int32_t aacls(xar_file_t f, const char *file) {
+static int32_t aacls(xar_t x, xar_file_t f, const char *file) {
 #ifdef HAVE_SYS_ACL_H
 #if !defined(__APPLE__)
 	acl_t a;
@@ -128,6 +140,9 @@ static int32_t aacls(xar_file_t f, const char *file) {
 	if( !type || (strcmp(type, "symlink") == 0) )
 		return 0;
 
+	if( !xar_check_prop(x, "acl") )
+		return 0;
+
 	a = acl_get_file(file, ACL_TYPE_DEFAULT);
 	if( a ) {
 		char *t;
@@ -168,6 +183,9 @@ DONE:
 	acl_t a;
 	int i;
 
+	if( !xar_check_prop(x, "acl") )
+		return 0;
+
 	a = acl_get_file(file, ACL_TYPE_EXTENDED);
 	if( !a )
 		return 0;
@@ -183,6 +201,8 @@ DONE:
 	}
 	acl_free(a);
 #endif /* !__APPLE__ */
+#else
+	(void)x; (void)f; (void)file;
 #endif
 	return 0;
 }
@@ -264,6 +284,8 @@ static int32_t eacls(xar_t x, xar_file_t f, const char *file) {
 		}
 	}
 #endif /* !__APPLE__ */
+#else
+	(void)x; (void)f; (void)file;
 #endif
 	return 0;
 }
@@ -279,10 +301,13 @@ static void x_addflag(xar_file_t f, const char *name) {
 }
 #endif
 
-static int32_t flags_archive(xar_file_t f, const struct stat *sb) {
+static int32_t flags_archive(xar_t x, xar_file_t f, const struct stat *sb) {
 #ifdef HAVE_STRUCT_STAT_ST_FLAGS
 	if( !sb->st_flags )
 		return 0;
+	
+	if( !xar_check_prop(x, XAR_FLAG_FORK) )
+		return 0;
 #ifdef UF_NODUMP
 	if( sb->st_flags & UF_NODUMP )
 		x_addflag(f, "UserNoDump");
@@ -299,6 +324,22 @@ static int32_t flags_archive(xar_file_t f, const struct stat *sb) {
 	if( sb->st_flags & UF_OPAQUE )
 		x_addflag(f, "UserOpaque");
 #endif
+#ifdef UF_NOUNLINK
+	if( sb->st_flags & UF_NOUNLINK )
+		x_addflag(f, "UserNoUnlink");
+#endif
+#ifdef UF_COMPRESSED
+	if( sb->st_flags & UF_COMPRESSED )
+		x_addflag(f, "UserCompressed");
+#endif
+#ifdef UF_TRACKED
+	if( sb->st_flags & UF_TRACKED )
+		x_addflag(f, "UserTracked");
+#endif
+#ifdef UF_HIDDEN
+	if( sb->st_flags & UF_HIDDEN )
+		x_addflag(f, "UserHidden");
+#endif
 #ifdef SF_ARCHIVED
 	if( sb->st_flags & SF_ARCHIVED )
 		x_addflag(f, "SystemArchived");
@@ -311,8 +352,18 @@ static int32_t flags_archive(xar_file_t f, const struct stat *sb) {
 	if( sb->st_flags & SF_APPEND )
 		x_addflag(f, "SystemAppend");
 #endif
+#ifdef SF_NOUNLINK
+	if( sb->st_flags & SF_NOUNLINK )
+		x_addflag(f, "SystemNoUnlink");
+#endif
+#ifdef SF_SNAPSHOT
+	if( sb->st_flags & SF_NOUNLINK )
+		x_addflag(f, "SystemSnapshot");
+#endif
 
-#endif /* HAVE_STRUCT_STAT_ST_FLAGS */
+#else /* !HAVE_STRUCT_STAT_ST_FLAGS */
+	(void)x; (void)f; (void)sb;
+#endif /* !HAVE_STRUCT_STAT_ST_FLAGS */
 	return 0;
 }
 
@@ -330,6 +381,11 @@ int32_t xar_flags_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 	char *tmp;
 	u_int flags = 0;
 
+	(void)buffer; (void)len;
+	/* data buffers cannot store flags information */
+	if (len)
+		return 0;
+
 	if( xar_prop_get(f, XAR_FLAG_FORK, NULL) )
 		return 0;
 
@@ -349,6 +405,34 @@ int32_t xar_flags_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 	if( x_getprop(f, "UserOpaque", (char **)&tmp) == 0 )
 		flags |= UF_OPAQUE;
 #endif
+#ifdef UF_NOUNLINK
+	if( x_getprop(f, "UserNoUnlink", (char **)&tmp) == 0 )
+		flags |= UF_NOUNLINK;
+#endif
+#ifdef UF_COMPRESSED
+#ifdef __APPLE__
+/* The UF_COMPRESSED attribute can only be restored by reapplying HFS+
+ * file compression (which is not yet supported) which will then set the bit.
+ * But, if a compressed file was archived from a system prior to 10.6 and we've
+ * grabbed a com.apple.decmpfs attribute, then we should set the bit on
+ * extraction because the file data is already HFS+ compressed and failing to
+ * set the bit would result in an unusable file.
+ */
+	if (x_getprop(f, "UserCompressed", (char **)&tmp) == 0 && xar_ea_find(f, "com.apple.decmpfs"))
+		flags |= UF_COMPRESSED;
+#else
+	if( x_getprop(f, "UserCompressed", (char **)&tmp) == 0 )
+		flags |= UF_COMPRESSED;
+#endif
+#endif
+#ifdef UF_TRACKED
+	if( x_getprop(f, "UserTracked", (char **)&tmp) == 0 )
+		flags |= UF_TRACKED;
+#endif
+#ifdef UF_HIDDEN
+	if( x_getprop(f, "UserHidden", (char **)&tmp) == 0 )
+		flags |= UF_HIDDEN;
+#endif
 #ifdef SF_ARCHIVED
 	if( x_getprop(f, "SystemArchived", (char **)&tmp) == 0 )
 		flags |= SF_ARCHIVED;
@@ -361,6 +445,14 @@ int32_t xar_flags_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 	if( x_getprop(f, "SystemAppend", (char **)&tmp) == 0 )
 		flags |= SF_APPEND;
 #endif
+#ifdef SF_NOUNLINK
+	if( x_getprop(f, "SystemNoUnlink", (char **)&tmp) == 0 )
+		flags |= SF_NOUNLINK;
+#endif
+#ifdef SF_SNAPSHOT
+	if( x_getprop(f, "SystemSnapshot", (char **)&tmp) == 0 )
+		flags |= SF_APPEND;
+#endif
 
 	if( !flags )
 		return 0;
@@ -375,6 +467,8 @@ int32_t xar_flags_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 		xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
 		return -1;
 	}
+#else
+	(void)x; (void)f; (void)file; (void)buffer; (void)len;
 #endif
 	
 	return 0;
@@ -388,9 +482,11 @@ int32_t xar_stat_archive(xar_t x, xar_file_t f, const char *file, const char *bu
 	struct tm t;
 	const char *type;
 
+	(void)buffer;
 	/* no stat attributes for data from a buffer, it is just a file */
 	if(len){
-		xar_prop_set(f, "type", "file");
+		if( xar_check_prop(x, "type") )
+			xar_prop_set(f, "type", "file");
 		return 0;
 	}
 	
@@ -405,21 +501,24 @@ int32_t xar_stat_archive(xar_t x, xar_file_t f, const char *file, const char *bu
 			return -1;
 		}
 		tmpf = xar_link_lookup(x, XAR(x)->sbcache.st_dev, XAR(x)->sbcache.st_ino, f);
-		xar_prop_set(f, "type", "hardlink");
-		if( tmpf ) {
-			const char *id;
-			id = xar_attr_get(tmpf, NULL, "id");
-			xar_attr_set(f, "type", "link", id);
-		} else {
-			xar_attr_set(f, "type", "link", "original");
+		if( xar_check_prop(x, "type") ) {
+			xar_prop_set(f, "type", "hardlink");
+			if( tmpf ) {
+				const char *id;
+				id = xar_attr_get(tmpf, NULL, "id");
+				xar_attr_set(f, "type", "link", id);
+			} else {
+				xar_attr_set(f, "type", "link", "original");
+			}
 		}
 	} else {
 		type = filetype_name(XAR(x)->sbcache.st_mode & S_IFMT);
-		xar_prop_set(f, "type", type);
+		if( xar_check_prop(x, "type") )
+			xar_prop_set(f, "type", type);
 	}
 
 	/* Record major/minor device node numbers */
-	if( S_ISBLK(XAR(x)->sbcache.st_mode) || S_ISCHR(XAR(x)->sbcache.st_mode)) {
+	if( xar_check_prop(x, "device") && (S_ISBLK(XAR(x)->sbcache.st_mode) || S_ISCHR(XAR(x)->sbcache.st_mode))) {
 		uint32_t major, minor;
 		char tmpstr[12];
 		xar_devmake(XAR(x)->sbcache.st_rdev, &major, &minor);
@@ -436,7 +535,8 @@ int32_t xar_stat_archive(xar_t x, xar_file_t f, const char *file, const char *bu
 		struct stat lsb;
 
 		memset(link, 0, sizeof(link));
-		readlink(file, link, sizeof(link)-1);
+		if (readlink(file, link, sizeof(link)-1) == -1)
+			return -1;
 		xar_prop_set(f, "link", link);
 		if( stat(file, &lsb) != 0 ) {
 			xar_attr_set(f, "link", "type", "broken");
@@ -446,47 +546,77 @@ int32_t xar_stat_archive(xar_t x, xar_file_t f, const char *file, const char *bu
 		}
 	}
 
-	asprintf(&tmpstr, "%04o", XAR(x)->sbcache.st_mode & (~S_IFMT));
-	xar_prop_set(f, "mode", tmpstr);
-	free(tmpstr);
+	if( xar_check_prop(x, "inode") ) {
+		if (asprintf(&tmpstr, "%"INO_STRING, XAR(x)->sbcache.st_ino) == -1)
+			return -1;
+		xar_prop_set(f, "inode", tmpstr);
+		free(tmpstr);
+	}
 
-	asprintf(&tmpstr, "%"PRIu64, (uint64_t)XAR(x)->sbcache.st_uid);
-	xar_prop_set(f, "uid", tmpstr);
-	free(tmpstr);
+	if( xar_check_prop(x, "deviceno") ) {
+		if (asprintf(&tmpstr, "%"DEV_STRING, DEV_CAST XAR(x)->sbcache.st_dev) == -1)
+			return -1;
+		xar_prop_set(f, "deviceno", tmpstr);
+		free(tmpstr);
+	}
 
-	pw = getpwuid(XAR(x)->sbcache.st_uid);
-	if( pw )
-		xar_prop_set(f, "user", pw->pw_name);
+	if( xar_check_prop(x, "mode") ) {
+		if (asprintf(&tmpstr, "%04o", XAR(x)->sbcache.st_mode & (~S_IFMT)) == -1)
+			return -1;
+		xar_prop_set(f, "mode", tmpstr);
+		free(tmpstr);
+	}
 
-	asprintf(&tmpstr, "%"PRIu64, (uint64_t)XAR(x)->sbcache.st_gid);
-	xar_prop_set(f, "gid", tmpstr);
-	free(tmpstr);
+	if( xar_check_prop(x, "uid") ) {
+		if (asprintf(&tmpstr, "%"PRIu64, (uint64_t)XAR(x)->sbcache.st_uid) == -1)
+			return -1;
+		xar_prop_set(f, "uid", tmpstr);
+		free(tmpstr);
+	}
 
-	gr = getgrgid(XAR(x)->sbcache.st_gid);
-	if( gr )
-		xar_prop_set(f, "group", gr->gr_name);
+	if( xar_check_prop(x, "user") ) {
+		pw = getpwuid(XAR(x)->sbcache.st_uid);
+		if( pw )
+			xar_prop_set(f, "user", pw->pw_name);
+	}
 
-	gmtime_r(&XAR(x)->sbcache.st_atime, &t);
-	memset(time, 0, sizeof(time));
-	strftime(time, sizeof(time), "%FT%T", &t);
-	strcat(time, "Z");
-	xar_prop_set(f, "atime", time);
+	if( xar_check_prop(x, "gid") ) {
+		if (asprintf(&tmpstr, "%"PRIu64, (uint64_t)XAR(x)->sbcache.st_gid) == -1)
+			return -1;
+		xar_prop_set(f, "gid", tmpstr);
+		free(tmpstr);
+	}
 
-	gmtime_r(&XAR(x)->sbcache.st_mtime, &t);
-	memset(time, 0, sizeof(time));
-	strftime(time, sizeof(time), "%FT%T", &t);
-	strcat(time, "Z");
-	xar_prop_set(f, "mtime", time);
+	if( xar_check_prop(x, "group") ) {
+		gr = getgrgid(XAR(x)->sbcache.st_gid);
+		if( gr )
+			xar_prop_set(f, "group", gr->gr_name);
+	}
 
-	gmtime_r(&XAR(x)->sbcache.st_ctime, &t);
-	memset(time, 0, sizeof(time));
-	strftime(time, sizeof(time), "%FT%T", &t);
-	strcat(time, "Z");
-	xar_prop_set(f, "ctime", time);
+	if( xar_check_prop(x, "atime") ) {
+		gmtime_r(&XAR(x)->sbcache.st_atime, &t);
+		memset(time, 0, sizeof(time));
+		strftime(time, sizeof(time), "%Y-%m-%dT%H:%M:%SZ", &t);
+		xar_prop_set(f, "atime", time);
+	}
 
-	flags_archive(f, &(XAR(x)->sbcache));
+	if( xar_check_prop(x, "mtime") ) {
+		gmtime_r(&XAR(x)->sbcache.st_mtime, &t);
+		memset(time, 0, sizeof(time));
+		strftime(time, sizeof(time), "%Y-%m-%dT%H:%M:%SZ", &t);
+		xar_prop_set(f, "mtime", time);
+	}
 
-	aacls(f, file);
+	if( xar_check_prop(x, "ctime") ) {
+		gmtime_r(&XAR(x)->sbcache.st_ctime, &t);
+		memset(time, 0, sizeof(time));
+		strftime(time, sizeof(time), "%Y-%m-%dT%H:%M:%SZ", &t);
+		xar_prop_set(f, "ctime", time);
+	}
+
+	flags_archive(x, f, &(XAR(x)->sbcache));
+
+	aacls(x, f, file);
 
 	return 0;
 }
@@ -500,7 +630,9 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 	struct tm t;
 	enum {ATIME=0, MTIME};
 	struct timeval tv[2];
+	char savesuid = 0;
 
+	(void)buffer;
 	/* when writing to a buffer, there are no permissions to set */
 	if ( len )
 		return 0;
@@ -528,6 +660,7 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 				g = gr->gr_gid;
 			}
 		}
+		savesuid = 1;
 	}
 	if( opt && (strcmp(opt, XAR_OPT_VAL_NUMERIC) == 0) ) {
 		xar_prop_get(f, "uid", &opt);
@@ -549,8 +682,13 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 			}
 			g = (gid_t)tmp;
 		}
+		savesuid = 1;
 	}
 
+	opt = xar_opt_get(x, XAR_OPT_SAVESUID);
+	if( opt && (strcmp(opt, XAR_OPT_VAL_TRUE) == 0) ) {
+		savesuid = 1;
+	}
 
 	xar_prop_get(f, "mode", &opt);
 	if( opt ) {
@@ -560,6 +698,10 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 			return -1;
 		}
 		m = (mode_t)tmp;
+		if( !savesuid ) {
+			m &= ~S_ISUID;
+			m &= ~S_ISGID;
+		}
 		mset = 1;
 	}
 
@@ -581,6 +723,8 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 			xar_err_set_file(x, f);
 			xar_err_set_string(x, "perm: could not lchown symlink");
 			xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
+			m &= ~S_ISUID;
+			m &= ~S_ISGID;
 		}
 #ifdef HAVE_LCHMOD
 		if( mset )
@@ -599,6 +743,8 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 			xar_err_set_string(x, "perm: could not chown file");
 			xar_err_set_errno(x, errno);
 			xar_err_callback(x, XAR_SEVERITY_NONFATAL, XAR_ERR_ARCHIVE_EXTRACTION);
+			m &= ~S_ISUID;
+			m &= ~S_ISGID;
 		}
 		if( mset )
 			if( chmod(file, m) ) {
@@ -615,7 +761,7 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 	xar_prop_get(f, "atime", &timestr);
 	if( timestr ) {
 		memset(&t, 0, sizeof(t));
-		strptime(timestr, "%FT%T", &t);
+		strptime(timestr, "%Y-%m-%dT%H:%M:%S", &t);
 		tv[ATIME].tv_sec = timegm(&t);
 	} else {
 		tv[ATIME].tv_sec = time(NULL);
@@ -624,7 +770,7 @@ int32_t xar_set_perm(xar_t x, xar_file_t f, const char *file, char *buffer, size
 	xar_prop_get(f, "mtime", &timestr);
 	if( timestr ) {
 		memset(&t, 0, sizeof(t));
-		strptime(timestr, "%FT%T", &t);
+		strptime(timestr, "%Y-%m-%dT%H:%M:%S", &t);
 		tv[MTIME].tv_sec = timegm(&t);
 	} else {
 		tv[MTIME].tv_sec = time(NULL);
@@ -639,6 +785,11 @@ int32_t xar_stat_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 	int ret, fd;
 	mode_t modet = 0;
 
+	(void)buffer;
+	/* data buffers cannot store stat information */
+	if (len)
+		return 0;
+
 	xar_prop_get(f, "type", &opt);
 	if(opt && (strcmp(opt, "character special") == 0))
 		modet = S_IFCHR;
@@ -656,7 +807,7 @@ int32_t xar_stat_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 			return -1;
 		if( (tmpll < 0) || (tmpll > 255) )
 			return -1;
-		major = tmpll;
+		major = (uint32_t)tmpll;
 
 		xar_prop_get(f, "device/minor", &opt);
 		tmpll = strtoll(opt, NULL, 10);
@@ -664,7 +815,7 @@ int32_t xar_stat_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 			return -1;
 		if( (tmpll < 0) || (tmpll > 255) )
 			return -1;
-		minor = tmpll;
+		minor = (uint32_t)tmpll;
 		
 		devt = xar_makedev(major, minor);
 		unlink(file);
@@ -724,7 +875,7 @@ int32_t xar_stat_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 			if( errno == ENOENT ) {
 				xar_iter_t i;
 				const char *ptr;
-				i = xar_iter_new(x);
+				i = xar_iter_new();
 				for(ptr = xar_prop_first(tmpf, i); ptr; ptr = xar_prop_next(i)) {
 					xar_iter_t a;
 					const char *val = NULL;
@@ -736,7 +887,7 @@ int32_t xar_stat_extract(xar_t x, xar_file_t f, const char *file, char *buffer,
 						continue;
 	
 					xar_prop_set(f, ptr, val);
-					a = xar_iter_new(x);
+					a = xar_iter_new();
 					for(akey = xar_attr_first(tmpf, ptr, a); akey; akey = xar_attr_next(a)) {
 						aval = xar_attr_get(tmpf, ptr, akey);
 						xar_attr_set(f, ptr, akey, aval);
